# This software is released under the Revised BSD License.
# See LICENSE for details
#
# Copyright (c) 2019, Ryan Chapin, https//:www.ryanchapin.com
# All rights reserved.

import os
from typing import Dict

class EnvVars(object):

    @staticmethod
    def get_env_vars(prefix: str) -> Dict:
        retval = {}
        for k, v in os.environ.items():
            if prefix in k:
                retval[k] = v
        return retval



from flask import Flask
from typing import Any, Callable, Iterable, Sequence
from prometheus_client import Counter, Gauge, Histogram, Metric, generate_latest, start_http_server
from rsyncdirector.lib.utils import Utils
import logging
import time

NAME = "metrics"

RUNS_COMPLETED = Counter("runs_completed", "Number of runs completed", labelnames=["rsync_id"])

JOB_DURATION = Histogram(
    name="job_duration_seconds",
    documentation="Duration of job in seconds",
    labelnames=["job_id"],
    buckets=(
        0.01,
        0.1,
        0.5,
        1,
        10,
        30,
        60,
        300,
        600,
        1800,
        3600,
        7200,
        14400,
        28800,
        86400,
        172800,
        604800,
        1209600,
    ),
)

JOB_SKIPPED_FOR_BLOCK_TIMEOUT_COUNTER = Counter(
    "job_skipped_for_block_timeout",
    "Number of times a job is skipped because a block timedout",
    labelnames=["job_id"],
)

BLOCKED_COUNTER = Counter("blocked", "Number of times a job is blocked", labelnames=["job_id"])

BLOCKED_DURATION = Histogram(
    "blocked_seconds",
    "Time blocked in seconds",
    labelnames=["job_id"],
    buckets=(
        0.01,
        0.1,
        0.5,
        1,
        10,
        30,
        60,
        300,
        600,
        1800,
        3600,
        7200,
        14400,
        28800,
        86400,
        172800,
        604800,
        1209600,
    ),
)

LOCK_FILES = Gauge("lock_files", "Number of currently existing lock files", labelnames=["job_id"])


class Metrics(object):
    def __init__(self, logger: logging.Logger, addr: str, port: str) -> None:
        self.logger = logger
        self.addr = addr
        self.port = port
        self.app = Flask(NAME)
        self.thread = None
        self.running = False

    def index(self):
        return "Metrics server is running"

    def metrics(self):
        return generate_latest(), 200, {"Content-Type": "text/plain; charset=utf-8"}

    def start(self, timeout: float, wait_time: float, num_retries: int):
        if not self.running:
            self.running = True
            self.prometheus_httpd, self.prometheus_thread = start_http_server(
                port=int(self.port), addr=self.addr
            )

            # We use "localhost" for the host because it is possible that we are configured to
            # listen on "0.0.0.0" and while that is a valid configuration to listen on a socket, it
            # is NOT valid for a URL.
            Utils.wait_for_http_service(
                logger=self.logger,
                host="localhost",
                port=int(self.port),
                path="/metrics",
                wait_time=wait_time,
                num_retries=num_retries,
                timeout=timeout,
            )

            self.logger.info(f"Metrics http server started on http://{self.addr}:{self.port}")
        else:
            self.logger.info(
                f"Metrics http server already running on http://{self.addr}:{self.port}"
            )

    def stop(self):
        if self.running:
            self.prometheus_httpd.shutdown()
            self.prometheus_httpd.server_close()
            self.prometheus_thread.join()
            self.running = False
            self.logger.info("Metrics http server stopped")
        else:
            self.logger.info("Metrics http is not running")



# This software is released under the Revised BSD License.
# See LICENSE for details
#
# Copyright (c) 2019, Ryan Chapin, https//:www.ryanchapin.com
# All rights reserved.

import sys
import logging
import signal
import time
from rsyncdirector.lib.rsyncdirector import RsyncDirector
from threading import Lock

# For the time-being, we are just logging to the console
logging.basicConfig(
    format="%(asctime)s,%(levelname)s,%(module)s,[%(threadName)s],%(message)s",
    level=logging.INFO,
    stream=sys.stdout,
)

logger = logging.getLogger(__name__)
rsyncdirector = None


def signal_handler(signal_number, _frame):
    global rsyncdirector

    if signal_number == signal.SIGHUP:
        logger.info(f"Executing run_once job after catching signal; signal_number={signal_number}")
        rsyncdirector.schedule_runonce_job()
    else:
        logger.info(f"Shutting down after catching signal; signal_number={signal_number}")
        rsyncdirector.shutdown()


def main():
    # Register signal handlers to properly shutdown the application.
    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGHUP, signal_handler)

    global rsyncdirector
    try:
        rsyncdirector = RsyncDirector(logger)
        rsyncdirector.start()

        while not rsyncdirector.is_shutdown():
            time.sleep(0.5)

    except Exception as e:
        # Dump the entire stack trace as we do not expect this case.
        logger.exception(e)

    if rsyncdirector:
        rsyncdirector.join(timeout=5.0)
    logger.info("Exiting main")


###############################################################################
# MAIN
###############################################################################

if __name__ == "__main__":
    main()



import logging
import sys
from dataclasses import dataclass, field
from rsyncdirector.lib.config import JobType
from invoke import run


class Rsync(object):
    def __init__(self, logger, type, sync, user=None, host=None, port=None, private_key_path=None):
        self.logger = logger
        self.type = type
        self.sync = sync
        self.user = user
        self.host = host
        self.port = port
        self.private_key_path = private_key_path

    def run(self):
        options = self.sync["opts"]
        remote_prefix = ""

        match self.type:
            case JobType.LOCAL:
                # Currently a noop as the command that we build is the proper one for local
                # synchronization.
                pass
            case JobType.REMOTE:
                remote_prefix = f"{self.user}@{self.host}:"

                ssh_opts = []
                if self.port:
                    ssh_opts.append(f"-p {self.port}")
                if self.private_key_path:
                    ssh_opts.append(f"-i {self.private_key_path}")
                if len(ssh_opts) > 0:
                    ssh_options = " ".join(ssh_opts)
                    options.extend(["-e", f"'ssh {ssh_options}'"])

            case _:
                raise Exception(f"invalid job type; self.type={self.type}")

        opts = " ".join(options)
        cmd = f"rsync {opts} {self.sync['source']} {remote_prefix}{self.sync['dest']}"
        self.logger.info(f"running rsync; cmd={cmd}")
        run(cmd)



import logging
import requests
import socket
import subprocess
import time
from typing import Tuple


class Utils(object):

    @staticmethod
    def run_bash_cmd(cmd: str, timeout_seconds: int = None) -> Tuple[int, str, str]:
        result = subprocess.run(
            cmd, shell=True, timeout=timeout_seconds, capture_output=True, executable="/bin/bash"
        )

        stdout = result.stdout.decode("ascii").strip()
        stderr = result.stderr.decode("ascii").strip()
        return result.returncode, stdout, stderr

    @staticmethod
    def wait_for_http_service(
        logger: logging.Logger,
        host: str,
        port: int,
        path: str,
        wait_time: int,
        num_retries: int,
        timeout: int = 1,
    ) -> None:
        num_tries = 0
        while True:
            response = None
            try:
                if num_tries >= num_retries:
                    raise Exception(
                        f"exhausted number of retries waiting for http service to accept connections; host={host}, port={port}"
                    )
                num_tries += 1

                url = f"http://{host}:{port}{path}"
                response = requests.get(url)
                response.raise_for_status()
                if response.status_code == 200:
                    return
                else:
                    logger.info(
                        f"waiting to retry for tcp port to accept connections; wait_time={wait_time}, host={host}, port={port}: {e}"
                    )
                    time.sleep(wait_time)

            except Exception as e:
                logger.info(
                    f"waiting for http service to accept connections; host={host}, port={port}: {e}"
                )
            finally:
                if response is not None:
                    response.close()



import logging
import os
import sys
import psutil
from abc import ABC, abstractmethod
from pathlib import Path
from fabric import Connection
from typing import Tuple


class PidFile(ABC):

    def __init__(self, logger: logging.Logger, pid: int, path: str) -> None:
        self.logger = logger
        self.pid = pid
        self.path = path

    @abstractmethod
    def delete(self) -> bool:
        pass

    @abstractmethod
    def does_pid_file_exist(self) -> Tuple[bool, int]:
        pass

    @abstractmethod
    def does_process_with_pid_exist(self) -> bool:
        pass

    @abstractmethod
    def write(self) -> bool:
        file_exists, existing_pid_str = self.does_pid_file_exist()
        existing_pid = None
        if existing_pid_str != "":
            try:
                existing_pid = int(existing_pid_str)
            except Exception as e:
                # If this fails, we will assume that there isn't a valid process running
                self.logger.warning(
                    f"pid file contained unparseable int; self.path={self.path}, existing_pid={existing_pid}"
                )
                existing_pid = None

        should_write_pid = False
        if file_exists and existing_pid:
            # Check to see if, for some reason that this is already our pid.
            if existing_pid == self.pid:
                should_write_pid = True
                self.logger.info(
                    f"existing local pid file contains our pid, overwriting it; path={self.path}, pid={self.pid}"
                )
            else:
                """
                If the existing pid is not ours, we need to verify that there is a running process
                with the pid that is in the current pid file.
                """
                if not self.does_process_with_pid_exist(existing_pid):
                    self.logger.info(
                        f"found existing pid file with existing pid that maps to non-existant process; path={self.path}, existing_pid={existing_pid}"
                    )
                    should_write_pid = True

        else:
            # There is no existing file or there is no valid pid in that file.
            should_write_pid = True

        return should_write_pid


class PidFileLocal(PidFile):

    def __init__(self, logger: logging.Logger, pid: int, path: str) -> None:
        super().__init__(logger, pid, path)

    def does_pid_file_exist(self) -> Tuple[bool, str]:
        path = Path(self.path)
        if path.is_dir():
            self.logger.warning(
                f"pid path is a directory for some reason, deleting it; self.path={self.path}"
            )
            path.rmdir()
            return False, ""

        file_exists = None
        existing_pid = ""
        if path.is_file():
            # Read the existing file and get the pid if there is one
            file_exists = True
            existing_pid = ""
            with open(self.path, "r") as fh:
                existing_pid = fh.read().strip()

        return file_exists, existing_pid

    def does_process_with_pid_exist(self, pid: int) -> bool:
        return psutil.pid_exists(pid)

    def delete(self) -> bool:
        os.remove(self.path)
        return True

    def write(self) -> bool:
        should_write_pid = super().write()

        if should_write_pid:
            with open(self.path, "w") as fh:
                bytes_written = fh.write(str(self.pid))
                if bytes_written > 0:
                    return True

        return False


class PidFileRemote(PidFile):

    def __init__(self, logger: logging.Logger, pid: int, path: str, conn: Connection) -> None:
        super().__init__(logger, pid, path)
        self.conn = conn

    def does_pid_file_exist(self) -> Tuple[bool, str]:
        result = self.conn.run(
            f'if [ -f "{self.path}" ]; then echo "1"; else echo "0"; fi', warn=True, hide=True
        )
        if not result.ok:
            self.logger.error(f"checking for existing pid file; path={self.path}, result={result}")
            return False, 0
        stdout = result.stdout
        file_exists = True if stdout.strip() == "1" else False

        if file_exists is False:
            return False, 0

        result = self.conn.run(f"cat {self.path}", warn=True, hide=True)
        if not result.ok:
            self.logger.error(f"cat'ing pid file path; path={self.path}, result={result}")
            return False, 0

        existing_pid = result.stdout
        return True, existing_pid.strip()

    def does_process_with_pid_exist(self, pid: int) -> bool:
        result = self.conn.run(f"ps -p {pid}", warn=True, hide=True)
        if not result.ok:
            # There is no process with this pid
            return False
        return True

    def delete(self) -> bool:
        result = self.conn.run(f"rm -f {self.path}")
        if not result.ok:
            self.logger.fatal(f"deleting pid file; path={self.path}, result={result}")
            sys.exit(1)
        return True

    def write(self) -> bool:
        should_write_pid = super().write()
        if should_write_pid:
            result = self.conn.run(f"echo {self.pid} > {self.path}")
            if not result.ok:
                self.logger.error(
                    f"writing remote pid file; pid={self.pid}, path={self.path}, result={result}"
                )
                return False
            return True
